;;; Guile-LZMA
;;; Copyright © 2021 Timothy Sample <samplet@ngyro.com>
;;;
;;; This file was adapted from the test file of Guile-zstd (version
;;; 0.1.1), with the following copyright notices:
;;;
;;; Copyright © 2019 Pierre Neidhardt <mail@ambrevar.xyz>
;;; Copyright © 2020 Mathieu Othacehe <othacehe@gnu.org>
;;; Copyright © 2020 Ludovic Courtès <ludo@gnu.org>
;;;
;;; This file is part of Guile-LZMA.
;;;
;;; Guile-LZMA is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; Guile-LZMA is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with Guile-LZMA.  If not, see <http://www.gnu.org/licenses/>.

(define-module (test-lzma)
  #:use-module (lzma)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-11)
  #:use-module (srfi srfi-34)
  #:use-module (srfi srfi-64)
  #:use-module (rnrs bytevectors)
  #:use-module (rnrs io ports)
  #:use-module (ice-9 match))

(define (random-seed)
  (logxor (getpid) (car (gettimeofday))))

(define %seed
  (let ((seed (random-seed)))
    (format (current-error-port) "random seed for tests: ~a~%"
            seed)
    (seed->random-state seed)))

(define (random-bytevector n)
  "Return a random bytevector of N bytes."
  (let ((bv (make-bytevector n)))
    (let loop ((i 0))
      (if (< i n)
          (begin
            (bytevector-u8-set! bv i (random 256 %seed))
            (loop (1+ i)))
          bv))))

(define* (compressed-data data #:key (level %default-compression-level)
                          extreme?
                          (check LZMA_CHECK_CRC64))
  (let-values (((port get) (open-bytevector-output-port)))
    (call-with-xz-output-port port
      (lambda (port)
        (put-bytevector port data))
      #:level level
      #:extreme? extreme?
      #:check check)
    (get)))

(define (compressed-data-check bv)
  (bit-extract (bytevector-u8-ref bv 7) 0 4))

(define* (compress-and-decompress data
                                  #:key (level %default-compression-level)
                                  extreme?)
  (bytevector=? (let ((compressed (compressed-data data #:level level
                                                   #:extreme? extreme?)))
                  (call-with-xz-input-port
                      (open-bytevector-input-port compressed)
                    (lambda (port)
                      (match (get-bytevector-all port)
                        ((? eof-object?) #vu8())
                        (bv bv)))))
                data))

(test-begin "lzma")

(test-assert "empty bytevector"
  (compress-and-decompress #vu8()))

(test-assert "random bytevector"
  (compress-and-decompress (random-bytevector (+ (random 100000)
                                                 (* 20 1024)))))
(test-assert "small bytevector"
  (compress-and-decompress (random-bytevector 127)))

(test-assert "one byte"
  (compress-and-decompress (random-bytevector 1)))

(test-assert "bytevector of 1MiB"
  (compress-and-decompress (random-bytevector (* 1024 1024))))

(test-assert "bytevector of 1MiB-1"
  (compress-and-decompress (random-bytevector (1- (* 1024 1024)))))

(test-assert "bytevector of 1MiB+1"
  (compress-and-decompress (random-bytevector (1+ (* 1024 1024)))))

(test-assert "bytevector of 2MiB, all compression levels"
  (let ((data (random-bytevector (* 2 1024 1024))))
    (every (lambda (level)
             (and (compress-and-decompress data #:level level
                                           #:extreme? #f)
                  (compress-and-decompress data #:level level
                                           #:extreme? #t)))
           (iota 9 1))))

(test-equal "truncated compressed stream"
  'LZMA_BUF_ERROR
  (let* ((compressed (compressed-data (random-bytevector 7777)))
         (size       (- (bytevector-length compressed) 142))
         (truncated  (make-bytevector size)))
    (bytevector-copy! compressed 0 truncated 0 size)
    (guard (exc
            ((lzma-error? exc) (lzma-error-symbol exc))
            (else #f))
      (call-with-xz-input-port (open-bytevector-input-port truncated)
        get-bytevector-all)
      #t)))

(test-equal "corrupt compressed stream"
  'LZMA_DATA_ERROR
  (let ((compressed (compressed-data (random-bytevector 7777))))
    (bytevector-u8-set! compressed 42
                        (logxor (bytevector-u8-ref compressed 42)
                                #xff))
    (guard (exc
            ((lzma-error? exc) (lzma-error-symbol exc))
            (else #f))
      (call-with-xz-input-port (open-bytevector-input-port compressed)
        get-bytevector-all)
      #t)))

(test-equal "compresses with the 'none' checksum"
  LZMA_CHECK_NONE
  (let ((compressed (compressed-data (random-bytevector (random 65536))
                                     #:check LZMA_CHECK_NONE)))
    (compressed-data-check compressed)))

(test-equal "compresses with the 'crc32' checksum"
  LZMA_CHECK_CRC32
  (let ((compressed (compressed-data (random-bytevector (random 65536))
                                     #:check LZMA_CHECK_CRC32)))
    (compressed-data-check compressed)))

(test-equal "compresses with the 'crc64' checksum"
  LZMA_CHECK_CRC64
  (let ((compressed (compressed-data (random-bytevector (random 65536))
                                     #:check LZMA_CHECK_CRC64)))
    (compressed-data-check compressed)))

(test-equal "compresses with the 'sha256' checksum"
  LZMA_CHECK_SHA256
  (let ((compressed (compressed-data (random-bytevector (random 65536))
                                     #:check LZMA_CHECK_SHA256)))
    (compressed-data-check compressed)))

(test-end)
